from pathlib import Path
import uuid

from flask import Flask
from flask_flatpages import Page
from htmd import site
from htmd.utils import set_post_metadata, validate_post
import pytest


def test_set_post_metadata_with_field_in_title(flask_app: Flask) -> None:
    # Add draft to the title
    example_path = Path('posts') / 'example.md'
    with example_path.open('r') as example_file:
        lines = example_file.readlines()

    title_line = 'title: Player drafted\n'
    with example_path.open('w') as example_file:
        for line in lines:
            if 'title:' in line:
                example_file.write(title_line)
            else:
                example_file.write(line)

    post = site.posts.get('example')
    post.meta['draft'] = 'build|' + str(uuid.uuid4())
    set_post_metadata(flask_app, post, 'draft', post.meta['draft'])

    with example_path.open('r') as example_file:
        contents = example_file.read()

    expected = 'draft: ' + post.meta['draft']
    assert expected in contents
    assert title_line in contents


def test_validate_post_invalid_updated(
    flask_app: Flask, # noqa: ARG001
    capsys: pytest.CaptureFixture[str],
) -> None:
    invalid_date = '2025-12-3'
    post = Page(
        path='example',
        meta=f'updated: {invalid_date}',
        body='',
        html_renderer=None,
        folder='.',
    )
    is_valid = validate_post(post, [])
    assert not is_valid
    exp = f'Updated date {invalid_date} for example is not in the format YYYY-MM-DD.\n'
    assert exp == capsys.readouterr().err



def test_validate_post_invalid_draft_uuid(
    flask_app: Flask, # noqa: ARG001
    capsys: pytest.CaptureFixture[str],
) -> None:
    invalid_draft = 'build|2025-12-03'
    post = Page(
        path='example',
        meta=f'draft: {invalid_draft}',
        body='',
        html_renderer=None,
        folder='.',
    )
    is_valid = validate_post(post, [])
    assert not is_valid
    expected = f'Draft field {invalid_draft} for example has an invalid UUID4.\n'
    assert expected == capsys.readouterr().err


def test_validate_post_invalid_draft_value(
    flask_app: Flask, # noqa: ARG001
    capsys: pytest.CaptureFixture[str],
) -> None:
    invalid_draft = 'foo'
    post = Page(
        path='example',
        meta=f'draft: {invalid_draft}',
        body='',
        html_renderer=None,
        folder='.',
    )
    is_valid = validate_post(post, [])
    assert not is_valid
    expected = 'Draft field foo for example is not valid. It must be True, False, "build", or "build|<UUID4>".\n' # noqa: E501
    assert expected == capsys.readouterr().err



from click.testing import CliRunner
from htmd.cli import templates


def test_templates(run_start: CliRunner) -> None:
    expected_output = (
        'templates/404.html was created.\n'
        'templates/_layout.html already exists and was not created.\n'
        'templates/_list.html was created.\n'
        'templates/all_posts.html was created.\n'
        'templates/all_tags.html was created.\n'
        'templates/author.html was created.\n'
        'templates/day.html was created.\n'
        'templates/index.html was created.\n'
        'templates/month.html was created.\n'
        'templates/post.html was created.\n'
        'templates/tag.html was created.\n'
        'templates/year.html was created.\n'
    )
    result = run_start.invoke(templates)
    assert result.exit_code == 0
    assert result.output == expected_output


def test_templates_without_folder() -> None:
    expected_output = 'templates/ directory not found.\n'
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(templates)
    assert result.exit_code == 1
    assert result.stderr == expected_output



import base64

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa


def encrypt_text(post_text: str) -> str:
    message = post_text.encode('utf-8')

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )
    
    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )
    pem_str = pem.decode('utf-8')
    pem_str = pem_str.replace('-----BEGIN PRIVATE KEY-----\n', '').replace('-----END PRIVATE KEY-----\n', '')

    public_key = private_key.public_key()

    ciphertext = public_key.encrypt(
        message,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None
        )
    )
    ciphertext_base64 = base64.b64encode(ciphertext).decode('utf-8')
    return pem_str, ciphertext_base64



from pathlib import Path
import shutil
import subprocess
import sys
import time
from types import TracebackType

from click.testing import CliRunner
from htmd.cli import PostsCreatedHandler, preview, StaticHandler
import pytest
import requests
from watchdog.events import DirCreatedEvent, FileCreatedEvent, FileModifiedEvent

from utils import set_example_to_draft, set_example_to_draft_build


def invoke_preview(run_start: CliRunner, args: list[str] | None = None) -> None:
    """
    run_start.invoke(preview) fails but it is used to track test coverage.

    I get a message that the path I pass to pytest is not found.
    I've track it down to this subprocess call.
    https://github.com/pallets/werkzeug/blob/d3dd65a27388fbd39d146caacf2563639ba622f0/src/werkzeug/_reloader.py#L273
    str(args) is "['/path/to/venv/bin/python', '-m', 'pytest', 'tests']"
    ERROR: file or directory not found: tests
    Which is how I'm running my tests.
    If I pass tests/test_preview.py then I see
    ERROR: file or directory not found: tests/test_preview.py
    """
    run_start.invoke(preview, args)


class run_preview:  # noqa: N801
    def __init__(
        self: 'run_preview',
        args: list[str] | None = None,
        max_tries: int = 10,
    ) -> None:
        self.args = args
        self.max_tries = max_tries

    def __enter__(self: 'run_preview') -> None:
        cmd = [sys.executable, '-m', 'htmd', 'preview']
        if self.args:
            cmd += self.args
        self.task = subprocess.Popen(cmd)  # noqa: S603
        url = 'http://localhost:9090/'
        count = 0
        while count < self.max_tries:  # pragma: no branch
            try:
                requests.get(url, timeout=1)
            except requests.exceptions.ConnectionError:
                count += 1
                time.sleep(0.1)
            else:
                break

    def __exit__(
        self: 'run_preview',
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.task.terminate()


def test_preview(run_start: CliRunner) -> None:  # noqa: ARG001
    url = 'http://localhost:9090/'
    with pytest.raises(requests.exceptions.ConnectionError):
        requests.get(url, timeout=1)
    success = 200
    with run_preview():
        response = requests.get(url, timeout=0.01)
        assert response.status_code == success


def test_preview_css_minify_js_minify(run_start: CliRunner) -> None:
    args = ['--css-minify', '--js-minify']
    invoke_preview(run_start, args)
    urls = (
        (200, 'http://localhost:9090/static/combined.min.css'),
        (200, 'http://localhost:9090/static/combined.min.js'),
    )
    js_path = Path('static') / 'scripts.js'
    with js_path.open('w') as js_file:
        js_file.write('document.getElementsByTagName("body");')

    with run_preview(args):
        for status, url in urls:
            response = requests.get(url, timeout=0.1)
            assert response.status_code == status


def test_preview_no_css_minify(run_start: CliRunner) -> None:
    args = ['--no-css-minify', '--js-minify']
    invoke_preview(run_start, args)
    urls = (
        (404, 'http://localhost:9090/static/combined.min.css'),
        (200, 'http://localhost:9090/static/combined.min.js'),
    )
    js_path = Path('static') / 'scripts.js'
    with js_path.open('w') as js_file:
        js_file.write('document.getElementsByTagName("body");')

    with run_preview(args):
        for status, url in urls:
            response = requests.get(url, timeout=0.1)
            assert response.status_code == status


def test_preview_css_minify_no_js_minify(run_start: CliRunner) -> None:
    args = ['--css-minify', '--no-js-minify']
    invoke_preview(run_start, args)
    urls = (
        (200, 'http://localhost:9090/static/combined.min.css'),
        (404, 'http://localhost:9090/static/combined.min.js'),
    )
    with run_preview(args):
        for status, url in urls:
            response = requests.get(url, timeout=0.1)
            assert response.status_code == status


def test_preview_no_css_minify_no_js_minify(run_start: CliRunner) -> None:
    args = ['--no-css-minify', '--no-js-minify']
    invoke_preview(run_start, args)
    urls = (
        (404, 'http://localhost:9090/static/combined.min.css'),
        (404, 'http://localhost:9090/static/combined.min.js'),
    )
    with run_preview(args):
        for status, url in urls:
            response = requests.get(url, timeout=0.1)
            assert response.status_code == status


@pytest.mark.parametrize('static_dir', [
    'static',
    'foo',
])
def test_preview_css_changes(run_start: CliRunner, static_dir: str) -> None:  # noqa: ARG001
    if static_dir != 'static':
        # Change static directory in config.toml
        config_path = Path('config.toml')
        with config_path.open('r') as config_file:
            lines = config_file.readlines()

        with config_path.open('w') as config_file:
            for line in lines:
                if 'static = ' in line:
                    config_file.write(f'static = "{static_dir}"\n')
                else:
                    config_file.write(line)

        # Ensure directory exists
        Path(static_dir).mkdir(exist_ok=True)

    url = 'http://localhost:9090/static/combined.min.css'
    new_style = 'p {color: red;}'
    expected = new_style.replace(' ', '').replace(';', '')
    with run_preview():
        response = requests.get(url, timeout=0.01)
        before = response.text
        assert expected not in before

        css_path = Path(static_dir) / 'style.css'
        # When static_dir != 'static' this will be a new file
        if static_dir != 'static':
            assert css_path.exists() is False

        with css_path.open('a') as css_file:
            css_file.write('\n' + expected + '\n')

        # Ensure new style is served
        read_timeout = False
        after = before
        max_attempts = 50_000
        attempts = 1

        # CSS changes can be seen without stopping webserver
        # Require two responses to be the same because
        # response was missing the last few characters
        # AssertionError: assert 'p{color:red}' in '\np{color:red'
        previous_response = None
        consecutive_same_responses = 0
        same_response_goal = 2
        while (
            (before == after or consecutive_same_responses < same_response_goal)
            and attempts < max_attempts
        ):
            try:
                response = requests.get(url, timeout=0.1)
            except (
                requests.exceptions.ChunkedEncodingError,
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
            ):  # pragma: no cover
                # Can happen when file is being replaced
                # Verify webserver didn't restart
                try:
                    response = requests.get('http://localhost:9090', timeout=0.1)
                except (
                    requests.exceptions.ChunkedEncodingError,
                    requests.exceptions.ConnectionError,
                    requests.exceptions.ReadTimeout,
                ):  # pragma: no cover
                    read_timeout = True  # pragma: no cover
            else:
                after = response.text

                if previous_response == after:
                    consecutive_same_responses += 1
                else:
                    consecutive_same_responses = 0
                previous_response = after
            attempts += 1

    assert read_timeout is False, 'Preview did reload.'
    assert before != after
    assert expected in after


@pytest.mark.parametrize('static_dir', [
    'static',
    'foo',
])
def test_preview_js_changes(run_start: CliRunner, static_dir: str) -> None:  # noqa: ARG001
    js_path = Path(static_dir) / 'script.js'
    if static_dir != 'static':
        # Change static directory in config.toml
        config_path = Path('config.toml')
        with config_path.open('r') as config_file:
            lines = config_file.readlines()

        with config_path.open('w') as config_file:
            for line in lines:
                if 'static = ' in line:
                    config_file.write(f'static = "{static_dir}"\n')
                else:
                    config_file.write(line)

        # Ensure directory exists
        Path(static_dir).mkdir(exist_ok=True)

        # Create file to test modified files update
        with js_path.open('w') as js_file:
            js_file.write('document.getElementByTagName("div")')

    url = 'http://localhost:9090/static/combined.min.js'
    expected = 'document.getElementByTagName("body")'

    with run_preview():
        response = requests.get(url, timeout=0.01)
        before = response.text
        assert expected not in before

        if static_dir != 'static':
            assert js_path.exists()

        with js_path.open('w') as js_file:
            js_file.write(expected)

        # Ensure new script is served
        read_timeout = False
        after = before
        max_attempts = 50_000
        attempts = 1

        # JS changes can be seen without stopping webserver
        # Require two responses to be the same because
        # response was missing the last few characters
        # after == 'document.getElementByTagName("body"' # noqa: ERA001
        previous_response = None
        consecutive_same_responses = 0
        same_response_goal = 2
        while (
            (before == after or consecutive_same_responses < same_response_goal)
            and attempts < max_attempts
        ):
            try:
                response = requests.get(url, timeout=0.1)
            except (
                requests.exceptions.ChunkedEncodingError,
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
            ):  # pragma: no cover
                # Can happen when file is being replaced
                # Verify webserver didn't restart
                try:
                    response = requests.get('http://localhost:9090', timeout=0.1)
                except (
                    requests.exceptions.ChunkedEncodingError,
                    requests.exceptions.ConnectionError,
                    requests.exceptions.ReadTimeout,
                ):  # pragma: no cover
                    read_timeout = True  # pragma: no cover
            else:
                after = response.text

                if previous_response == after:
                    consecutive_same_responses += 1
                else:
                    consecutive_same_responses = 0
                previous_response = after

            attempts += 1

    assert read_timeout is False, 'Preview did reload.'
    assert before != after
    assert expected in after


@pytest.mark.parametrize('posts_dir', [
    'posts',
    'bar',
])
def test_preview_when_posts_change(run_start: CliRunner, posts_dir: str) -> None:  # noqa: ARG001
    if posts_dir != 'posts':
        # Change static directory in config.toml
        config_path = Path('config.toml')
        with config_path.open('r') as config_file:
            lines = config_file.readlines()

        with config_path.open('w') as config_file:
            for line in lines:
                if 'posts = ' in line:
                    config_file.write(f'posts = "{posts_dir}"\n')
                else:
                    config_file.write(line)

        # Ensure directory exists
        Path(posts_dir).mkdir(exist_ok=True)

    url = 'http://localhost:9090/'
    title = 'Test Title'
    expected = 'This is the content.'
    with run_preview():
        response = requests.get(url, timeout=0.01)
        before = response.text
        assert expected not in before
        assert title not in before

        # file will be new when posts_dir != 'posts'
        post_path = Path(posts_dir) / 'example.md'
        with post_path.open('w') as post_file:
            post_file.write('---\n')
            post_file.write(f'title: {title}\n')
            post_file.write('published: 2025-11-05\n')
            post_file.write('...\n')
            post_file.write(f'{expected}\n')

        # Ensure new sentence is available after reload
        read_timeout = False
        after = before
        max_attempts = 50_000
        attempts = 1
        while after == before and attempts < max_attempts:
            try:
                response = requests.get(url, timeout=0.1)
            except (
                requests.exceptions.ChunkedEncodingError,
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
            ):  # pragma: no cover
                # happens during restart
                read_timeout = True
            else:
                after = response.text

            attempts += 1

        assert read_timeout is False, 'Preview did reload.'
        assert before != after
        assert expected in after
        assert title in after


@pytest.mark.parametrize('pages_dir', [
    'pages',
    'bar',
])
def test_preview_shows_pages_change_without_reload(
    run_start: CliRunner,  # noqa: ARG001
    pages_dir: str,
) -> None:
    if pages_dir != 'pages':
        # Change static directory in config.toml
        config_path = Path('config.toml')
        with config_path.open('r') as config_file:
            lines = config_file.readlines()

        with config_path.open('w') as config_file:
            for line in lines:
                if 'pages = ' in line:
                    config_file.write(f'pages = "{pages_dir}"\n')
                else:
                    config_file.write(line)

        # Ensure directory exists
        Path(pages_dir).mkdir(exist_ok=True)
        # Move example page into new pages folder
        shutil.copy(Path('pages') / 'about.html', Path(pages_dir))

    url = 'http://localhost:9090/about/'
    expected = 'This is new.'
    page_path = Path(pages_dir) / 'about.html'
    with page_path.open('r') as page_file:
        contents = page_file.read()

    contents = contents.replace(
        '</p>',
        f' {expected}</p>',
    )
    with run_preview():
        response = requests.get(url, timeout=0.01)
        before = response.text
        assert expected not in before

        with page_path.open('w') as page_file:
            page_file.write(contents)

        # Ensure new sentence is available after change
        read_timeout = False
        after = before
        max_attempts = 50_000
        attempts = 1

        # Since HTML changes can be seen without reloading
        while before == after and attempts < max_attempts:
            try:
                response = requests.get(url, timeout=0.1)
            except (
                requests.exceptions.ChunkedEncodingError,
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
            ):  # pragma: no cover
                # happens during restart
                read_timeout = True  # pragma: no cover
            else:
                after = response.text

            attempts += 1

        assert read_timeout is False, 'Preview did reload.'
        assert before != after
        assert expected in after


def test_preview_shows_new_pages(run_start: CliRunner) -> None:  # noqa: ARG001
    page_path = Path('pages') / 'about.html'
    with page_path.open('r') as page_file:
        contents = page_file.read()

    new_path_path = Path('pages') / 'new.html'
    expected = 'This is new.'
    contents = contents.replace(
        '</p>',
        f' {expected}</p>',
    )
    url = 'http://localhost:9090/new/'
    with run_preview():
        response = requests.get(url, timeout=0.01)
        before = response.text
        assert response.status_code == 404  # noqa: PLR2004
        assert expected not in before

        with new_path_path.open('w') as page_file:
            page_file.write(contents)

        # Ensure new sentence is available after change
        read_timeout = False
        after = before
        max_attempts = 50_000
        attempts = 1

        # Since HTML changes can be seen without reloading
        while before == after and attempts < max_attempts:
            try:
                response = requests.get(url, timeout=0.1)
            except (
                requests.exceptions.ChunkedEncodingError,
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
            ):  # pragma: no cover
                # happens during restart
                read_timeout = True  # pragma: no cover
            else:
                after = response.text

            attempts += 1

        assert read_timeout is False, 'Preview did reload.'
        assert before != after
        assert expected in after


def test_preview_drafts(run_start: CliRunner) -> None:
    args = ['--drafts']
    invoke_preview(run_start, args)
    set_example_to_draft()
    success = 200

    urls = (
        (404, '/2014/'),
        (404, '/2014/10/'),
        (404, '/2014/10/30/'),
        (404, '/2014/10/30/example/'),
        (404, '/tags/first/'),
        (404, '/author/Taylor/'),
    )
    not_in = (
        '/',
        '/all/',
    )
    # drafts should not appear
    with run_preview():
        for status, url in urls:
            response = requests.get('http://localhost:9090' + url, timeout=1)
            assert response.status_code == status

        for url in not_in:
            response = requests.get('http://localhost:9090' + url, timeout=0.01)
            assert response.status_code == success
            assert 'Example Post' not in response.text

    # drafts should appear
    with run_preview(args):
        for _status, url in urls:
            response = requests.get('http://localhost:9090' + url, timeout=1)
            assert response.status_code == success

        for url in not_in:
            response = requests.get('http://localhost:9090' + url, timeout=0.01)
            assert response.status_code == success
            assert 'Example Post' in response.text

    set_example_to_draft_build()
    urls = (
        (404, '/2014/'),
        (404, '/2014/10/'),
        (404, '/2014/10/30/'),
        (200, '/2014/10/30/example/'),
        (200, '/tags/first/'),
        (200, '/author/Taylor/'),
    )
    not_in = (
        '/',
        '/all/',
    )
    with run_preview():
        for status, url in urls:
            response = requests.get('http://localhost:9090' + url, timeout=1)
            assert response.status_code == status

        for url in not_in:
            response = requests.get('http://localhost:9090' + url, timeout=0.01)
            assert response.status_code == success
            assert 'Example Post' not in response.text


def test_preview_when_static_folder_does_not_exist(run_start: CliRunner) -> None:
    static_path = Path('static')
    for file_in_dir in static_path.iterdir():
        file_in_dir.unlink()

    static_path.rmdir()

    assert static_path.exists() is False

    # invoke_preview is only used for test coverage
    invoke_preview(run_start)

    url = 'http://localhost:9090/'
    success = 200
    with run_preview():
        assert static_path.exists() is False
        response = requests.get(url, timeout=0.01)
        assert response.status_code == success


def test_preview_when_combined_js_exists(run_start: CliRunner) -> None:
    combined_path = Path('static') / 'combined.min.js'

    with combined_path.open('w') as combined_js_file:
        combined_js_file.write('document.getElementByTagName("body");')

    new_path = Path('static') / 'new.js'
    new_js = 'console.log("new");'
    with new_path.open('w') as new_js_file:
        new_js_file.write(new_js)

    # invoke_preview is only used for test coverage
    invoke_preview(run_start)

    url = 'http://localhost:9090/static/combined.min.js'
    success = 200
    with run_preview():
        response = requests.get(url, timeout=0.01)
        assert response.status_code == success
        assert new_js in response.text

    # invoke again will exit combine_and_minify_js() early
    # since there is no change
    # invoke_preview is only used for test coverage
    invoke_preview(run_start)

    url = 'http://localhost:9090/static/combined.min.js'
    success = 200
    with run_preview():
        response = requests.get(url, timeout=0.01)
        assert response.status_code == success
        assert new_js in response.text


def test_static_handler(run_start: CliRunner) -> None:  # noqa: ARG001
    static_handler = StaticHandler(Path('static'))

    # Add new file to combined.min.css
    new_css = 'body { background-color: aqua;}'
    new_css_path = Path('static') / 'new.css'
    with new_css_path.open('w') as new_css_file:
        new_css_file.write(new_css)

    css_file_event = FileModifiedEvent(bytes(new_css_path), '', is_synthetic=True)
    static_handler.on_modified(css_file_event)

    new_js = 'document.getElementByTag("body")'
    new_js_path = Path('static') / 'new.js'
    with new_js_path.open('w') as new_js_file:
        new_js_file.write(new_js)
    js_file_event = FileModifiedEvent(str(new_js_path), '', is_synthetic=True)
    static_handler.on_modified(js_file_event)
    # Verify exit early when .js file event but no changes
    static_handler.on_modified(js_file_event)


def test_posts_handler(run_start: CliRunner) -> None:  # noqa: ARG001
    posts_handler = PostsCreatedHandler()

    # Add non .md file
    non_md_path = Path('posts') / 'not_markdown.txt'
    with non_md_path.open('w') as non_md_file:
        non_md_file.write('This is not markdown.')

    new_file_event = FileCreatedEvent(bytes(non_md_path), '', is_synthetic=True)
    posts_handler.on_created(new_file_event)

    posts_path = Path('posts')
    new_dir_event = DirCreatedEvent(bytes(posts_path), '', is_synthetic=True)
    posts_handler.on_created(new_dir_event)


def test_favicon(run_start: CliRunner) -> None:  # noqa: ARG001
    url = 'http://localhost:9090/static/favicon.svg'
    success = 200
    with run_preview():
        response = requests.get(url, timeout=0.1)
    assert response.status_code == success
    assert response.headers['Content-Type'] == 'image/svg+xml; charset=utf-8'
    assert len(response.content) > 0
    with (Path('static') / 'favicon.svg').open('rb') as favicon_file:
        svg_content = favicon_file.read()
    assert response.content == svg_content



from collections.abc import Generator
from pathlib import Path
import re

from click.testing import CliRunner
from htmd.cli import build, start
import pytest

from utils import (
  remove_fields_from_post,
  set_example_to_draft,
  set_example_to_draft_build,
  SUCCESS_REGEX,
)


def copy_example_as_draft_build() -> None:
    post_path = Path('posts') / 'example.md'
    copy_path = Path('posts') / 'copy.md'
    with post_path.open('r') as post_file:
        lines = post_file.readlines()
    with copy_path.open('w') as copy_file:
        for line in lines:
            if 'draft' in line:
                copy_file.write('draft: build\n')
            else:
                copy_file.write(line)


def get_draft_uuid(path: str) -> str:
    draft_path = Path('posts') / f'{path}.md'
    with draft_path.open('r') as draft_file:
        for line in draft_file.readlines():
            if 'draft: build' in line:
                return line.replace('draft: build|', '').strip()
    return ''


@pytest.fixture(scope='module')
def build_draft() -> Generator[CliRunner]:
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(start)
        set_example_to_draft()
        copy_example_as_draft_build()
        result = runner.invoke(build)
        assert result.exit_code == 0
        assert re.search(SUCCESS_REGEX, result.output)
        # Tests code is run here
        yield runner


def test_draft_only_draft_build_is_in_build(build_draft: CliRunner) -> None:
    post_path = Path('build') / '2014' / '10' / '30' / 'example' / 'index.html'
    assert post_path.exists() is False

    example_uuid = get_draft_uuid('example')
    assert example_uuid == ''
    draft_uuid = get_draft_uuid('copy')
    draft_path = Path('build') / 'draft' / draft_uuid / 'index.html'
    assert draft_path.is_file() is True

    # build again now that draft has uuid
    result = build_draft.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)
    assert draft_path.is_file() is True


def test_no_drafts_home(build_draft: CliRunner) -> None:  # noqa: ARG001
    with (Path('build') / 'index.html').open('r') as home_page:
        assert 'Example Post' not in home_page.read()


def test_no_drafts_atom_feed(build_draft: CliRunner) -> None:  # noqa: ARG001
    with (Path('build') / 'feed.atom').open('r') as feed_page:
        assert 'Example Post' not in feed_page.read()


def test_no_drafts_all_posts(build_draft: CliRunner) -> None:  # noqa: ARG001
    with (Path('build') / 'all' / 'index.html').open('r') as web_page:
        assert 'Example Post' not in web_page.read()


def test_no_drafts_all_tags(build_draft: CliRunner) -> None:  # noqa: ARG001
    with (Path('build') / 'tags' / 'index.html').open('r') as web_page:
        assert 'first' not in web_page.read()


def test_no_drafts_in_tag(build_draft: CliRunner) -> None:  # noqa: ARG001
    # tag page exists because the draft links to it
    with (Path('build') / 'tags' / 'first' / 'index.html').open('r') as web_page:
        assert 'Example Post' not in web_page.read()


def test_no_drafts_for_author(build_draft: CliRunner) -> None:  # noqa: ARG001
    # author page exists because the draft links to it
    with (Path('build') / 'author' / 'Taylor' / 'index.html').open('r') as web_page:
        assert 'Example Post' not in web_page.read()


def test_no_drafts_for_year(build_draft: CliRunner) -> None:  # noqa: ARG001
    # folder exists becaues of URL for post
    assert (Path('build') / '2014' / 'index.html').exists() is False


def test_no_drafts_for_month(build_draft: CliRunner) -> None:  # noqa: ARG001
    # folder exists becaues of URL for post
    assert (Path('build') / '2014' / '10' / 'index.html').exists() is False


def test_no_drafts_for_day(build_draft: CliRunner) -> None:  # noqa: ARG001
    # folder exists becaues of URL for post
    assert (Path('build') / '2014' / '10' / '30' / 'index.html').exists() is False


def test_draft_without_published(run_start: CliRunner) -> None:
    set_example_to_draft()
    remove_fields_from_post('example', ('published', 'updated'))
    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)
    draft_path = Path('build') / 'draft' / 'example' / 'index.html'
    assert not draft_path.is_file()


def test_draft_build_and_without_published(run_start: CliRunner) -> None:
    set_example_to_draft_build()
    remove_fields_from_post('example', ('published', 'updated'))
    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)
    draft_uuid = get_draft_uuid('example')
    draft_path = Path('build') / 'draft' / draft_uuid / 'index.html'
    assert draft_path.is_file() is True



from pathlib import Path

from click.testing import CliRunner
from htmd.cli import start


def test_start() -> None:
    runner = CliRunner()
    expected_output = (
        'templates/ was created.\n'
        'templates/_layout.html was created.\n'
        'static/ was created.\n'
        'static/_reset.css was created.\n'
        'static/style.css was created.\n'
        'static/favicon.svg was created.\n'
        'pages/ was created.\n'
        'pages/about.html was created.\n'
        'posts/ was created.\n'
        'posts/example.md was created.\n'
        'config.toml was created.\n'
        'Add the site name and edit settings in config.toml\n'
    )
    with runner.isolated_filesystem():
        result = runner.invoke(start)

        expected_files = (
            'config.toml',
            'pages/about.html',
            'posts/example.md',
            'static/_reset.css',
            'static/style.css',
            'static/favicon.svg',
            'templates/_layout.html',
        )
        expected_dirs = (
            'pages',
            'posts',
            'static',
            'templates',
        )

        for file in expected_files:
            file_path = Path(file)
            assert file_path.exists()
            assert file_path.is_file()

        for folder in expected_dirs:
            dir_path = Path(folder)
            assert dir_path.exists()

    assert result.exit_code == 0
    assert result.output == expected_output


def test_start_all_templates() -> None:
    runner = CliRunner()
    expected_output = (
        'templates/ was created.\n'
        'templates/404.html was created.\n'
        'templates/_layout.html was created.\n'
        'templates/_list.html was created.\n'
        'templates/all_posts.html was created.\n'
        'templates/all_tags.html was created.\n'
        'templates/author.html was created.\n'
        'templates/day.html was created.\n'
        'templates/index.html was created.\n'
        'templates/month.html was created.\n'
        'templates/post.html was created.\n'
        'templates/tag.html was created.\n'
        'templates/year.html was created.\n'
        'static/ was created.\n'
        'static/_reset.css was created.\n'
        'static/style.css was created.\n'
        'static/favicon.svg was created.\n'
        'pages/ was created.\n'
        'pages/about.html was created.\n'
        'posts/ was created.\n'
        'posts/example.md was created.\n'
        'config.toml was created.\n'
        'Add the site name and edit settings in config.toml\n'
    )
    with runner.isolated_filesystem():
        result = runner.invoke(start, ['--all-templates'])
        expected_files = (
            'config.toml',
            'pages/about.html',
            'posts/example.md',
            'static/_reset.css',
            'static/style.css',
            'static/favicon.svg',
            'templates/404.html',
            'templates/_layout.html',
            'templates/_list.html',
            'templates/all_posts.html',
            'templates/all_tags.html',
            'templates/author.html',
            'templates/day.html',
            'templates/index.html',
            'templates/month.html',
            'templates/post.html',
            'templates/tag.html',
            'templates/year.html',
        )
        expected_dirs = (
            'pages',
            'posts',
            'static',
            'templates',
        )
        for file in expected_files:
            file_path = Path(file)
            assert file_path.exists()
            assert file_path.is_file()

        for folder in expected_dirs:
            dir_path = Path(folder)
            assert dir_path.exists()

    assert result.exit_code == 0
    assert result.output == expected_output


def test_start_with_existing_template() -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path('templates').mkdir()
        layout_path = Path('templates') / '_layout.html'
        with layout_path.open('w') as layout:
            pass
        result = runner.invoke(start)
        with layout_path.open('r') as layout:
            # _layout.html was not replaced
            assert layout.read() == ''
    assert result.exit_code == 0
    assert 'templates/ already exists and was not created.' in result.output
    expected2 = 'templates/_layout.html already exists and was not created.'
    assert expected2 in result.output



import os
from pathlib import Path
import re
import shutil

from click.testing import CliRunner
from htmd.cli import build

from utils import remove_fields_from_post, set_example_to_draft, SUCCESS_REGEX


def test_build(run_start: CliRunner) -> None:
    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_verify_fails(run_start: CliRunner) -> None:
    expected_output = 'Post "example" does not have field title.\n'
    remove_fields_from_post('example', ('title',))
    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert result.output == expected_output


def test_build_js_minify(run_start: CliRunner) -> None:
    with (Path('static') / 'app.js').open('w') as js_file:
        js_file.write('console.log("htmd");')

    result = run_start.invoke(build, ['--js-minify'])
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)
    with (Path('build') / 'index.html').open('r') as built_index:
        contents = built_index.read()
    assert 'combined.min.js' in contents


def test_build_js_minify_no_js_files(run_start: CliRunner) -> None:
    result = run_start.invoke(build, ['--js-minify'])
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_no_js_minify(run_start: CliRunner) -> None:
    result = run_start.invoke(build, ['--no-js-minify'])
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_css_minify(run_start: CliRunner) -> None:
    result = run_start.invoke(build, ['--css-minify'])
    with (Path('build') / 'index.html').open('r') as built_index:
        contents = built_index.read()
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)
    assert 'combined.min.css' in contents
    # combined.min.css is used instead of style.css
    static_build_files = [
        file.name for file in (Path('build') / 'static').iterdir()]
    assert 'combined.min.css' in static_build_files
    assert 'style.css' not in static_build_files


def test_build_no_css_minify(run_start: CliRunner) -> None:
    result = run_start.invoke(build, ['--no-css-minify'])
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_css_minify_no_css_files(run_start: CliRunner) -> None:
    (Path('static') / 'style.css').unlink()
    (Path('static') / '_reset.css').unlink()
    result = run_start.invoke(build, ['--css-minify'])
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_no_css_minify_no_js_minify(run_start: CliRunner) -> None:
    result = run_start.invoke(build, ['--no-css-minify', '--no-js-minify'])
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_html_pretty_true(run_start: CliRunner) -> None:
    config_path = Path('config.toml')
    with config_path.open('r') as config_file:
        lines = config_file.readlines()

    with config_path.open('w') as config_file:
        for line in lines:
            if 'pretty =' in line:
                config_file.write('pretty = true\n')
            else:
                config_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_html_minify_true(run_start: CliRunner) -> None:
    config_path = Path('config.toml')
    with config_path.open('r') as config_file:
        lines = config_file.readlines()

    with config_path.open('w') as config_file:
        for line in lines:
            if 'minify =' in line:
                config_file.write('minify = true\n')
            else:
                config_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_page_404(run_start: CliRunner) -> None:
    # Linking to a page that doesn't exist
    # will cause a 404 status code
    # and stop the build
    expected_output = (
        'All posts are correctly formatted.\n'
        "Unexpected status '404 NOT FOUND' on URL /dne/\n"
    )
    about_path = Path('pages') / 'about.html'
    with about_path.open('r') as about_file:
        lines = about_file.readlines()

    new_line = '''<p><a href="{{ url_for('pages.page', path='dne') }}">DNE link</a></p>\n'''  # noqa: E501
    with about_path.open('w') as about_file:
        for line in lines:
            if '<p>This is the about page.</p>' in line:
                about_file.write(new_line)
            else:
                about_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert result.output == expected_output


def test_build_multiple_posts(run_start: CliRunner) -> None:
    shutil.copyfile(
        Path('posts') / 'example.md',
        Path('posts') / 'sample.md',
    )
    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_from_sub_directory(run_start: CliRunner) -> None:
    current_directory = Path.cwd()
    os.chdir(Path(current_directory) / 'posts')
    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_feed_dot_atom(run_start: CliRunner) -> None:
    result = run_start.invoke(build)
    assert result.exit_code == 0
    current_directory = Path.cwd()
    assert (Path(current_directory) / 'build' / 'feed.atom').is_file()


def test_build_page_without_link(run_start: CliRunner) -> None:
    page_lines = (
        "{% extends '_layout.html' %}\n",
        '\n',
        '{% block title %}New{% endblock title %}\n',
        '\n',
        '{% block content %}\n',
        '  <article>\n',
        '    <h1>New</h1>\n',
        '    <p>Totally new</p>\n',
        '  </article>\n',
        '{% endblock content %}\n',
    )
    # Create page that is doesn't have a link in the site
    with (Path('pages') / 'new.html').open('w') as page_file:
        for line in page_lines:
            page_file.write(line)
    result = run_start.invoke(build)
    assert result.exit_code == 0
    with (Path('build') / 'new' / 'index.html').open('r') as page_file:
        assert 'Totally new' in page_file.read()


def test_build_empty_directory() -> None:
    expected = 'Can not find config.toml\n'
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(build)
        assert result.exit_code == 1
        assert result.output == expected


def test_build_without_static(run_start: CliRunner) -> None:
    path = Path('static')
    shutil.rmtree(path)
    assert path.exists() is False
    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_without_posts(run_start: CliRunner) -> None:
    path = Path('posts')
    shutil.rmtree(path)
    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_without_pages(run_start: CliRunner) -> None:
    path = Path('pages')
    shutil.rmtree(path)

    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert "Unexpected status '404 NOT FOUND' on URL /about/" in result.output

    # Remove link from _layout.html
    layout_path = Path('templates') / '_layout.html'
    with layout_path.open('r') as layout_file:
        lines = layout_file.readlines()
    with layout_path.open('w') as layout_file:
        for line in lines:
            if 'about' in line:
                continue
            layout_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_without_templates(run_start: CliRunner) -> None:
    path = Path('templates')
    shutil.rmtree(path)
    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)


def test_build_drafts_removed_from_build(run_start: CliRunner) -> None:
    path_list = [
        Path('build') / '2014' / 'index.html',
        Path('build') / '2014' / '10' / 'index.html',
        Path('build') / '2014' / '10' / '30' / 'index.html',
        Path('build') / '2014' / '10' / '30' / 'example' / 'index.html',
    ]
    # create example post in build
    run_start.invoke(build)
    for path in path_list:
        assert path.is_file()

    set_example_to_draft()
    run_start.invoke(build)
    for path in path_list:
        assert path.is_file() is False


def test_build_vcs_repo(run_start: CliRunner) -> None:
    path_list = [
        Path('build') / '.git',
        Path('build') / '.hg',
    ]
    run_start.invoke(build)
    for path in path_list:
        path.mkdir()
        with (path / 'keep').open('w') as keep_file:
            keep_file.write('keep')

    run_start.invoke(build)
    for path in path_list:
        assert path.is_dir()


def test_build_with_default_author(run_start: CliRunner) -> None:
    config_path = Path('config.toml')
    with config_path.open('r') as config_file:
        lines = config_file.readlines()

    with config_path.open('w') as config_file:
        for line in lines:
            if 'default_name' in line:
                config_file.write('default_name = "Taylor"\n')
            else:
                config_file.write(line)

    remove_fields_from_post('example', ('draft',))

    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)
    post_path = Path('build') / '2014' / '10' / '30' / 'example' / 'index.html'
    with post_path.open('r') as post_file:
        contents = post_file.read()
    assert 'Posted by <a href="/author/Taylor/">Taylor</a> on 2014-10-30' in contents


def test_build_contains_favicon(run_start: CliRunner) -> None:
    with (Path('static') / 'favicon.svg').open('r') as favicon_file:
        original_contents = favicon_file.read()
    result = run_start.invoke(build)
    assert result.exit_code == 0
    assert re.search(SUCCESS_REGEX, result.output)
    build_favicon = Path('build') / 'static' / 'favicon.svg'
    assert build_favicon.is_file()
    with build_favicon.open('r') as favicon_file:
        contents = favicon_file.read()
    assert contents == original_contents



import datetime
from pathlib import Path
import signal
import sys
import threading
import types
import warnings

import click
from flask import Flask
from flask_flatpages import FlatPages
from watchdog.events import (
    DirCreatedEvent,
    DirModifiedEvent,
    FileCreatedEvent,
    FileModifiedEvent,
    FileSystemEvent,
    FileSystemEventHandler,
)
from watchdog.observers import Observer

from . import site
from .utils import (
    combine_and_minify_css,
    combine_and_minify_js,
    copy_missing_templates,
    copy_site_file,
    create_directory,
    send_stderr,
    set_post_metadata,
    validate_post,
)


warnings.filterwarnings('ignore', '.*Nothing frozen for endpoints.*')


@click.group()
@click.version_option()
def cli() -> None:
    pass  # pragma: no cover


@cli.command('start', short_help='Create example files to get started.')
@click.option(
    '--all-templates',
    is_flag=True,
    default=False,
    help='Include all templates.',
)
def start(all_templates: bool) -> None:  # noqa: FBT001
    dir_templates = create_directory('templates/')
    if all_templates:
        copy_missing_templates()
    else:
        copy_site_file(dir_templates, '_layout.html')

    dir_static = create_directory('static/')
    copy_site_file(dir_static, '_reset.css')
    copy_site_file(dir_static, 'style.css')
    copy_site_file(dir_static, 'favicon.svg')

    dir_pages = create_directory('pages/')
    copy_site_file(dir_pages, 'about.html')

    dir_posts = create_directory('posts/')
    copy_site_file(dir_posts, 'example.md')

    copy_site_file(Path(), 'config.toml')
    click.echo('Add the site name and edit settings in config.toml')


@cli.command('verify', short_help='Verify posts formatting is correct.')
def verify() -> Flask:
    app = site.init_app()

    correct = True
    required_fields = ['title']
    # Only check author if there is no default
    if not app.config.get('DEFAULT_AUTHOR'):
        required_fields.append('author')
    for post in site.posts:
        if not validate_post(post, required_fields):
            correct = False

    if correct:
        msg = 'All posts are correctly formatted.'
        click.echo(click.style(msg, fg='green'))

    # Check if SITE_NAME exists
    site_name = app.config.get('SITE_NAME')
    if not site_name:
        # SITE_NAME is not required
        message = '[site] name is not set in config.toml.'
        click.echo(click.style(message, fg='yellow'))

    if not correct:
        sys.exit(1)
    return app


def set_posts_datetime(app: Flask, posts: FlatPages) -> None:
    # Ensure each post has a published date
    # set time for correct date field
    for post in posts:
        if post.meta.get('draft', False):
            continue
        if 'updated' not in post.meta:
            published = post.meta.get('published')
            if isinstance(published, datetime.datetime):
                field = 'updated'
            else:
                field = 'published'
        else:
            field = 'updated'

        post_datetime = post.meta.get(field)
        now = datetime.datetime.now(tz=datetime.UTC)
        current_time = now.time()
        if isinstance(post_datetime, datetime.date):
            post_datetime = datetime.datetime.combine(
                post_datetime, current_time,
            ).replace(tzinfo=datetime.UTC)
        else:
            post_datetime = now
        post.meta[field] = post_datetime
        set_post_metadata(app, post, field, post_datetime.isoformat())


@cli.command('build', short_help='Create static version of the site.')
@click.pass_context
@click.option(
    '--css-minify/--no-css-minify',
    default=True,
    help='If CSS should be minified',
)
@click.option(
    '--js-minify/--no-js-minify',
    default=True,
    help='If JavaScript should be minified',
)
def build(
    ctx: click.Context,
    css_minify: bool,  # noqa: FBT001
    js_minify: bool,  # noqa: FBT001
) -> None:
    app = ctx.invoke(verify)
    # If verify fails sys.exit(1) will run

    assert app.static_folder is not None
    static_path = Path(app.static_folder)
    if css_minify and combine_and_minify_css(static_path):
        app.config['INCLUDE_CSS'] = app.jinja_env.globals['INCLUDE_CSS'] = True

    if js_minify and combine_and_minify_js(static_path):
        app.config['INCLUDE_JS'] = app.jinja_env.globals['INCLUDE_JS'] = True

    set_posts_datetime(app, site.posts)

    freezer = site.freezer
    try:
        freezer.freeze()
    except ValueError as exc:
        send_stderr(str(exc))
        sys.exit(1)

    build_dir = app.config.get('FREEZER_DESTINATION')
    msg = f'Static site was created in {build_dir}'
    click.echo(click.style(msg, fg='green'))


class StaticHandler(FileSystemEventHandler):
    def __init__(self, static_directory: Path) -> None:
        super().__init__()
        self.static_directory = static_directory

    def on_modified(self, event: DirModifiedEvent | FileModifiedEvent) -> None:
        if event.is_directory:
            return
        src_path = event.src_path
        if isinstance(src_path, bytes):
            src_path = src_path.decode('utf-8')
        dst_css = 'combined.min.css'
        dst_js = 'combined.min.js'
        if dst_css in src_path or dst_js in src_path or '.swp' in src_path:
            return

        if src_path.endswith('.css') and combine_and_minify_css(self.static_directory):
            click.echo(f'Changes in {src_path}. Recreating {dst_css}...')
        elif src_path.endswith('.js') and combine_and_minify_js(self.static_directory):
            click.echo(f'Changes in {src_path}. Recreating {dst_js}...')


class PostsCreatedHandler(FileSystemEventHandler):
    def handle_event(self, event: FileSystemEvent, is_new_post: bool) -> None:  # noqa: FBT001
        if event.is_directory:
            return
        src_path = event.src_path
        if isinstance(src_path, bytes):
            src_path = src_path.decode('utf-8')
        if not src_path.endswith('.md'):
            return

        site.reload_posts()
        for post in site.posts:
            validate_post(post, [])

        action = 'created' if is_new_post else 'updated'
        click.echo(f'Post {action} {src_path}.')

    def on_created(self, event: DirCreatedEvent | FileCreatedEvent) -> None:
        self.handle_event(event, is_new_post=True)

    def on_modified(self, event: DirModifiedEvent | FileModifiedEvent) -> None:
        self.handle_event(event, is_new_post=False)


def watch_disk(
    static_folder: str,
    posts_path: Path,
    exit_event: threading.Event,
) -> None:
    observer = Observer()

    static_directory = Path(static_folder)
    static_handler = StaticHandler(static_directory)
    observer.schedule(
        static_handler,
        path=str(static_directory),
        recursive=True,
    )

    posts_handler = PostsCreatedHandler()
    observer.schedule(
        posts_handler,
        path=str(posts_path),
        recursive=True,
    )

    observer.start()

    try:
        while not exit_event.is_set():
            observer.join(1)
    finally:
        observer.stop()
        observer.join()


@cli.command('preview', short_help='Serve files to preview site.')
@click.pass_context
@click.option(
    '--host', '-h',
    default='127.0.0.1',
    help='Location to access the files.',
)
@click.option(
    '--port', '-p',
    default=9090,
    help='Port on which to serve the files.',
)
@click.option(
    '--css-minify/--no-css-minify',
    default=True,
    help='If CSS should be minified',
)
@click.option(
    '--js-minify/--no-js-minify',
    default=True,
    help='If JavaScript should be minified',
)
@click.option(
    '--drafts',
    default=False,
    help='Show draft posts in the preview.',
    is_flag=True,
)
def preview(
    _ctx: click.Context,
    host: str,
    port: int,
    css_minify: bool,  # noqa: FBT001
    js_minify: bool,  # noqa: FBT001
    drafts: bool,  # noqa: FBT001
) -> None:
    app = site.init_app(drafts)

    assert site.app.static_folder is not None
    if css_minify and combine_and_minify_css(Path(site.app.static_folder)):
        app.config['INCLUDE_CSS'] = app.jinja_env.globals['INCLUDE_CSS'] = True

    if js_minify and combine_and_minify_js(Path(site.app.static_folder)):
        app.config['INCLUDE_JS'] = app.jinja_env.globals['INCLUDE_JS'] = True

    stop_event = threading.Event()

    def handle_sigterm(
        _signum: int,
        _frame: types.FrameType | None,
    ) -> None:  # pragma: no cover
        stop_event.set()
        raise KeyboardInterrupt

    signal.signal(signal.SIGTERM, handle_sigterm)
    signal.signal(signal.SIGINT, handle_sigterm)

    watch_thread = threading.Thread(
        target=watch_disk,
        args=(
            app.static_folder,
            app.config['FLATPAGES_ROOT'],
            stop_event,
        ),
    )
    watch_thread.start()

    try:
        app.run(debug=True, host=host, port=port)
    finally:
        # After Flask has been stopped stop watchdog
        stop_event.set()


@cli.command('templates', short_help='Create any missing templates')
def templates() -> None:
    try:
        copy_missing_templates()
    except FileNotFoundError:
        send_stderr('templates/ directory not found.')
        sys.exit(1)



from pathlib import Path

from click.testing import CliRunner
from htmd.cli import verify

from utils import remove_fields_from_post


def test_verify(run_start: CliRunner) -> None:
    result = run_start.invoke(verify)
    assert result.exit_code == 0
    expected_output = 'All posts are correctly formatted.\n'
    assert result.output == expected_output


def test_verify_author_missing(run_start: CliRunner) -> None:
    # Remove author from example post
    remove_fields_from_post('example', ('author',))

    result = run_start.invoke(verify)
    assert result.exit_code == 1
    expected_output = 'Post "example" does not have field author.\n'
    assert result.stderr == expected_output


def test_verify_title_missing(run_start: CliRunner) -> None:
    # Remove title from example post
    remove_fields_from_post('example', ('title',))

    result = run_start.invoke(verify)
    assert result.exit_code == 1
    expected_output = 'Post "example" does not have field title.\n'
    assert result.stderr == expected_output


def test_verify_published_missing(run_start: CliRunner) -> None:
    # Remove published from example post
    remove_fields_from_post('example', ('published',))

    result = run_start.invoke(verify)
    # verify doesn't check for published
    # since it will be added on build.
    expected_output = 'All posts are correctly formatted.\n'
    assert result.output == expected_output
    assert result.exit_code == 0


def test_verify_published_invalid_year(run_start: CliRunner) -> None:
    example_post_path = Path('posts') / 'example.md'
    with example_post_path.open('r') as post:
        lines = post.readlines()
    with example_post_path.open('w') as post:
        for line in lines:
            if 'published' in line:
                post.write('published: 14-10-30\n')
            else:
                post.write(line)

    result = run_start.invoke(verify)
    assert result.exit_code == 1
    expected_output = (
        'Published date 14-10-30 for example'
        ' is not in the format YYYY-MM-DD.\n'
    )
    assert result.output == expected_output


def test_verify_published_invalid_month(run_start: CliRunner) -> None:
    example_post_path = Path('posts') / 'example.md'
    with example_post_path.open('r') as post:
        lines = post.readlines()
    with example_post_path.open('w') as post:
        for line in lines:
            if 'published' in line:
                post.write('published: 2014-1-30\n')
            else:
                post.write(line)

    result = run_start.invoke(verify)
    assert result.exit_code == 1
    expected_output = (
        'Published date 2014-1-30 for example'
        ' is not in the format YYYY-MM-DD.\n'
    )
    assert result.output == expected_output


def test_verify_published_invalid_day(run_start: CliRunner) -> None:
    example_post_path = Path('posts') / 'example.md'
    with example_post_path.open('r') as post:
        lines = post.readlines()
    with example_post_path.open('w') as post:
        for line in lines:
            if 'published' in line:
                post.write('published: 2014-01-3\n')
            else:
                post.write(line)

    result = run_start.invoke(verify)
    assert result.exit_code == 1
    expected_output = (
        'Published date 2014-01-3 for example'
        ' is not in the format YYYY-MM-DD.\n'
    )
    assert result.output == expected_output


def test_verify_site_name_empty(run_start: CliRunner) -> None:
    expected_output = (
        'All posts are correctly formatted.\n'
        '[site] name is not set in config.toml.\n'
    )

    config_path = Path('config.toml')
    with config_path.open('r') as post:
        lines = post.readlines()
    with config_path.open('w') as post:
        seen = False
        for line in lines:
            if 'name' in line and not seen:
                # [site] name is the first name
                seen = True
                post.write("name = ''\n")
            else:
                post.write(line)

    result = run_start.invoke(verify)
    # [site] name isn't required
    assert result.exit_code == 0
    assert result.output == expected_output


def test_verify_site_name_missing(run_start: CliRunner) -> None:
    expected_output = (
        'All posts are correctly formatted.\n'
        '[site] name is not set in config.toml.\n'
    )
    config_path = Path('config.toml')
    with config_path.open('r') as post:
        lines = post.readlines()
    with config_path.open('w') as post:
        for line in lines:
            if 'name' not in line:
                post.write(line)

    result = run_start.invoke(verify)

    assert result.exit_code == 0
    assert result.output == expected_output


def test_verify_no_config(run_start: CliRunner) -> None:
    expected_output = 'Can not find config.toml\n'
    Path('config.toml').unlink()
    result = run_start.invoke(verify)

    assert result.exit_code == 1
    assert result.output == expected_output



import datetime
from pathlib import Path

from click.testing import CliRunner
from htmd.cli import build

from utils import remove_fields_from_post


def test_build_post_404_invalid_date_year(run_start: CliRunner) -> None:
    # Linking to a post with incorrect values
    # for dates will cause 404 and stop the build
    expected_output = (
        'All posts are correctly formatted.\n'
        "Unexpected status '404 NOT FOUND' on URL /14/10/30/example/\n"
    )

    about_path = Path('pages') / 'about.html'
    with about_path.open('r') as about_file:
        lines = about_file.readlines()

    new_line = '''<p><a href="{{ url_for('post', year=14, month='10', day='30', path='example') }}">DNE link</a></p>\n'''  # noqa: E501
    with about_path.open('w') as about_file:
        for line in lines:
            if '<p>This is the about page.</p>' in line:
                about_file.write(new_line)
            else:
                about_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert result.output == expected_output


def test_build_post_404_invalid_date_month(run_start: CliRunner) -> None:
    # Linking to a post with incorrect values
    # for dates will cause 404 and stop the build
    expected_output = (
        'All posts are correctly formatted.\n'
        "Unexpected status '404 NOT FOUND' on URL /2014/1/30/example/\n"
    )

    about_path = Path('pages') / 'about.html'
    with about_path.open('r') as about_file:
        lines = about_file.readlines()

    new_line = '''<p><a href="{{ url_for('post', year=2014, month='1', day='30', path='example') }}">DNE link</a></p>\n'''  # noqa: E501
    with about_path.open('w') as about_file:
        for line in lines:
            if '<p>This is the about page.</p>' in line:
                about_file.write(new_line)
            else:
                about_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert result.output == expected_output


def test_build_post_404_invalid_date_day(run_start: CliRunner) -> None:
    # Linking to a post with incorrect values
    # for dates will cause 404 and stop the build
    expected_output = (
        'All posts are correctly formatted.\n'
        "Unexpected status '404 NOT FOUND' on URL /2014/10/3/example/\n"
    )

    about_path = Path('pages') / 'about.html'
    with about_path.open('r') as about_file:
        lines = about_file.readlines()

    new_line = '''<p><a href="{{ url_for('post', year=2014, month='10', day='3', path='example') }}">DNE link</a></p>\n'''  # noqa: E501
    with about_path.open('w') as about_file:
        for line in lines:
            if '<p>This is the about page.</p>' in line:
                about_file.write(new_line)
            else:
                about_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert result.output == expected_output


def test_build_post_404_different_date(run_start: CliRunner) -> None:
    # Linking to a page with the wrong date
    # will cause a 404 status code
    # and stop the build
    expected_output = (
        'All posts are correctly formatted.\n'
        "Unexpected status '404 NOT FOUND' on URL /2014/10/29/example/\n"
    )
    about_path = Path('pages') / 'about.html'
    with about_path.open('r') as about_file:
        lines = about_file.readlines()

    new_line = '''<p><a href="{{ url_for('post', year=2014, month='10', day='29', path='example') }}">DNE link</a></p>\n'''  # noqa: E501
    with about_path.open('w') as about_file:
        for line in lines:
            if '<p>This is the about page.</p>' in line:
                about_file.write(new_line)
            else:
                about_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert result.output == expected_output


def test_build_year_404_incorrect(run_start: CliRunner) -> None:
    # Linking to a page with the wrong date
    # will cause a 404 status code
    # and stop the build
    expected_output = (
        'All posts are correctly formatted.\n'
        "Unexpected status '404 NOT FOUND' on URL /14/\n"
    )

    about_path = Path('pages') / 'about.html'
    with about_path.open('r') as about_file:
        lines = about_file.readlines()

    new_line = '''<p><a href="{{ url_for('year_view', year=14) }}">DNE link</a></p>\n'''
    with about_path.open('w') as about_file:
        for line in lines:
            if '<p>This is the about page.</p>' in line:
                about_file.write(new_line)
            else:
                about_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert result.output == expected_output


def test_build_year_404_no_posts(run_start: CliRunner) -> None:
    # Linking to a page with the wrong date
    # will cause a 404 status code
    # and stop the build
    expected_output = (
        'All posts are correctly formatted.\n'
        "Unexpected status '404 NOT FOUND' on URL /2013/\n"
    )
    about_path = Path('pages') / 'about.html'
    with about_path.open('r') as about_file:
        lines = about_file.readlines()

    new_line = '''<p><a href="{{ url_for('year_view', year=2013) }}">DNE link</a></p>\n'''  # noqa: E501
    with about_path.open('w') as about_file:
        for line in lines:
            if '<p>This is the about page.</p>' in line:
                about_file.write(new_line)
            else:
                about_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert result.output == expected_output


def test_build_month_404_no_posts(run_start: CliRunner) -> None:
    # Linking to a page with the wrong date
    # will cause a 404 status code
    # and stop the build
    expected_output = (
        'All posts are correctly formatted.\n'
        "Unexpected status '404 NOT FOUND' on URL /2014/01/\n"
    )

    about_path = Path('pages') / 'about.html'
    with about_path.open('r') as about_file:
        lines = about_file.readlines()

    new_line = '''<p><a href="{{ url_for('month_view', year=2014, month='01') }}">DNE link</a></p>\n'''  # noqa: E501
    with about_path.open('w') as about_file:
        for line in lines:
            if '<p>This is the about page.</p>' in line:
                about_file.write(new_line)
            else:
                about_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert result.output == expected_output


def test_build_day_404_no_posts(run_start: CliRunner) -> None:
    # Linking to a page with the wrong date
    # will cause a 404 status code
    # and stop the build
    expected_output = (
        'All posts are correctly formatted.\n'
        "Unexpected status '404 NOT FOUND' on URL /2014/10/29/\n"
    )

    about_path = Path('pages') / 'about.html'
    with about_path.open('r') as about_file:
        lines = about_file.readlines()

    new_line = '''<p><a href="{{ url_for('day_view', year=2014, month='10', day='29') }}">DNE link</a></p>\n'''  # noqa: E501
    with about_path.open('w') as about_file:
        for line in lines:
            if '<p>This is the about page.</p>' in line:
                about_file.write(new_line)
            else:
                about_file.write(line)

    result = run_start.invoke(build)
    assert result.exit_code == 1
    assert result.output == expected_output


def test_build_updated_time_is_added(run_start: CliRunner) -> None:
    # If there is no published/updated time then
    # build will add it
    # verify that time is not there
    # ensure correct time is added
    example_path = Path('posts') / 'example.md'
    with example_path.open('r') as post_file:
        b_lines = post_file.readlines()
    result = run_start.invoke(build)
    assert result.exit_code == 0
    with example_path.open('r') as post_file:
        a_lines = post_file.readlines()
    for b_line, a_line in zip(b_lines, a_lines, strict=True):
        if 'updated' in b_line:
            b_updated = b_line
            a_updated = a_line
        if 'published' in b_line:
            b_published = b_line
            a_published = a_line

    # Verify published didn't change
    assert a_published is not None
    assert b_published.strip() == a_published.strip()

    assert a_updated.startswith(b_updated.strip())
    assert len(a_updated) > len(b_updated)
    b_datetime_str = b_updated.replace('updated:', '').strip()
    a_datetime_str = a_updated.replace('updated:', '').strip()
    b_datetime = datetime.datetime.fromisoformat(b_datetime_str)
    a_datetime = datetime.datetime.fromisoformat(a_datetime_str)

    # Before didn't have a time
    assert b_datetime.hour == 0
    assert b_datetime.minute == 0
    assert b_datetime.second == 0

    date_with_current_time = datetime.datetime.now(tz=datetime.UTC).replace(
        year=a_datetime.year,
        month=a_datetime.month,
        day=a_datetime.day,
    )
    time_difference = abs(a_datetime - date_with_current_time)

    # Verify updated time is close to now
    threshold_seconds = 60
    assert time_difference.total_seconds() < threshold_seconds


def test_build_published_time_is_added(run_start: CliRunner) -> None:
    # If there is no published/updated time then
    # build will add it
    # verify that time is not there
    # ensure correct time is added
    remove_fields_from_post('example', ('updated',))
    example_path = Path('posts') / 'example.md'
    with example_path.open('r') as post_file:
        b_lines = post_file.readlines()
    result = run_start.invoke(build)
    assert result.exit_code == 0
    with example_path.open('r') as post_file:
        a_lines = post_file.readlines()
    for b_line, a_line in zip(b_lines, a_lines, strict=True):
        if 'published' in b_line:
            b_published = b_line
            a_published = a_line

    assert a_published.startswith(b_published.strip())
    assert len(a_published) > len(b_published)
    b_datetime_str = b_published.replace('published:', '').strip()
    a_datetime_str = a_published.replace('published:', '').strip()
    b_datetime = datetime.datetime.fromisoformat(b_datetime_str)
    a_datetime = datetime.datetime.fromisoformat(a_datetime_str)

    # Before didn't have a time
    assert b_datetime.hour == 0
    assert b_datetime.minute == 0
    assert b_datetime.second == 0

    date_with_current_time = datetime.datetime.now(datetime.UTC).replace(
        year=a_datetime.year,
        month=a_datetime.month,
        day=a_datetime.day,
    )
    time_difference = abs(a_datetime - date_with_current_time)

    # Verify published time is close to now
    threshold_seconds = 60
    assert time_difference.total_seconds() < threshold_seconds

    # verify updated is not added
    assert 'updated' not in ''.join(a_lines)


def test_build_updated_is_added(run_start: CliRunner) -> None:
    """
    Verify updated is added.

    If published has a time and there is no updated
    then build will add updated with time.
    """
    # Remove updated from example post
    remove_fields_from_post('example', ('updated',))
    # First build adds time to published
    result = run_start.invoke(build)
    assert result.exit_code == 0
    # Second build adds updated with time
    result2 = run_start.invoke(build)
    assert result2.exit_code == 0
    example_path = Path('posts') / 'example.md'
    with example_path.open('r') as post_file:
        a_lines = post_file.readlines()
    for a_line in a_lines:
        if 'updated' in a_line:
            a_updated = a_line

    a_datetime_str = a_updated.replace('updated:', '').strip()
    a_datetime = datetime.datetime.fromisoformat(a_datetime_str)

    date_with_current_time = datetime.datetime.now(datetime.UTC).replace(
        year=a_datetime.year,
        month=a_datetime.month,
        day=a_datetime.day,
    )
    time_difference = abs(a_datetime - date_with_current_time)

    # Verify published time is close to now
    threshold_seconds = 60
    assert time_difference.total_seconds() < threshold_seconds


def test_build_updated_is_added_once(run_start: CliRunner) -> None:
    # Remove updated from example post
    example_path = Path('posts') / 'example.md'
    with example_path.open('r') as post_file:
        b_lines = post_file.readlines()
    with example_path.open('w') as post_file:
        for line in b_lines:
            if 'updated' not in line:
                post_file.write(line)
        # add "..." to post content
        post_file.write('...\n')

    # First build adds published time
    result = run_start.invoke(build)
    assert result.exit_code == 0
    # Second build adds updated
    result2 = run_start.invoke(build)
    assert result2.exit_code == 0
    with example_path.open('r') as post_file:
        a_lines = post_file.readlines()
    count = 0
    for a_line in a_lines:
        if 'updated' in a_line:
            count += 1

    assert count == 1


def test_build_without_published(run_start: CliRunner) -> None:
    remove_fields_from_post('example', ('published', 'updated'))

    # First build adds published time
    result = run_start.invoke(build)
    assert result.exit_code == 0
    example_path = Path('posts') / 'example.md'
    with example_path.open('r') as post_file:
        a_lines = post_file.readlines()
    count = 0
    for a_line in a_lines:
        if 'published' in a_line:
            count += 1

    assert count == 1


def test_build_with_post_in_each_month(run_start: CliRunner) -> None:
    post_path = Path('posts') / 'example.md'
    with post_path.open('r') as post_file:
        lines = post_file.readlines()
    for month in range(1, 13):
        with post_path.open('w') as post_file:
            for line in lines:
                if 'published' in line:
                    post_file.write(f'published: 2014-{month:02}-03\n')
                else:
                    post_file.write(line)

        result = run_start.invoke(build)
        assert result.exit_code == 0



from flask import Flask
from flask.testing import FlaskClient
from htmd import site
import pytest

from utils import set_example_to_draft


@pytest.fixture
def client(flask_app: Flask) -> FlaskClient:
    return flask_app.test_client()


def test_author_does_not_exist(client: FlaskClient) -> None:
    # If the author doesn't exist it will be a 404
    response = client.get('/author/dne/')
    assert response.status_code == 404  # noqa: PLR2004


def test_page_does_not_exist(client: FlaskClient) -> None:
    # Ensure htmd preview matches build
    # Only pages will be served
    # before this change pages.page was serving templates
    response = client.get('/author/')
    assert response.status_code == 404  # noqa: PLR2004


def test_draft_does_not_exist(client: FlaskClient) -> None:
    # Ensure htmd preview matches build
    # Only pages will be served
    # before this change pages.page was serving templates
    response = client.get('/draft/dne/')
    assert response.status_code == 404  # noqa: PLR2004


def test_tag_does_not_exist(client: FlaskClient) -> None:
    found = 200
    not_found = 404
    # If the author doesn't exist it will be a 404
    response = client.get('/tags/dne/')
    assert response.status_code == not_found

    set_example_to_draft()
    response = client.get('/tags/first/')
    assert response.status_code == not_found
    response = client.get('/author/Taylor/')
    assert response.status_code == not_found
    response = client.get('/2014/10/30/example/')
    assert response.status_code == not_found

    site.reload_posts(show_drafts=True)
    response = client.get('/tags/first/')
    assert response.status_code == found
    response = client.get('/author/Taylor/')
    assert response.status_code == found



from collections.abc import Iterator
from pathlib import Path
import sys
import tomllib
import typing
import uuid

from bs4 import BeautifulSoup
from feedwerk.atom import AtomFeed
from flask import abort, Blueprint, Flask, render_template, Response, url_for
from flask.typing import ResponseReturnValue
from flask_flatpages import FlatPages, Page, pygments_style_defs
from flask_frozen import Freezer
from htmlmin import minify
from jinja2 import ChoiceLoader, FileSystemLoader

from .utils import set_post_metadata, valid_uuid


this_dir = Path(__file__).parent


class Posts(FlatPages):
    def __init__(self, app: Flask) -> None:
        super().__init__(app)
        self.show_drafts: bool = False
        self.published_posts: list[Page] = []


app = Flask(
    __name__,
    # default templates
    template_folder=this_dir / 'example_site' / 'templates',
)
freezer = Freezer(app, with_static_files=False)
pages = Blueprint('pages', __name__)
posts = Posts(app)


def get_project_dir() -> Path:
    current_directory = Path.cwd()

    while not (current_directory / 'config.toml').is_file():
        parent_directory = current_directory.parent

        if current_directory == parent_directory:
            return Path.cwd()

        current_directory = parent_directory

    return current_directory


def reload_posts(show_drafts: bool | None = None) -> None: # noqa: FBT001
    posts.reload()
    if show_drafts is not None:
        posts.show_drafts = show_drafts
    if posts.show_drafts:
        posts.published_posts = [
            p for p in posts
            if 'published' in p.meta and hasattr(p.meta['published'], 'year')
        ]
    else:
        posts.published_posts = [
            p for p in posts
            if (
                not p.meta.get('draft', False)
                and 'published' in p.meta
                and hasattr(p.meta['published'], 'year')
            )
        ]


def init_app(show_drafts: bool = False) -> Flask: # noqa: FBT001,FBT002
    project_dir = get_project_dir()

    try:
        with (project_dir / 'config.toml').open('rb') as config_file:
            htmd_config = tomllib.load(config_file)
    except FileNotFoundError:
        msg = 'Can not find config.toml'
        sys.exit(msg)

    # Flask configs are flat, config.toml is not
    # Define the configuration keys and their default values
    # 'Flask config': [section, key, default]
    config_keys : dict[str, tuple[str, str, typing.Any]] = {
        'SITE_NAME': ('site', 'name', ''),
        'SITE_URL': ('site', 'url', ''),
        'SITE_LOGO': ('site', 'logo', ''),
        'SITE_DESCRIPTION': ('site', 'description', ''),
        'SITE_TWITTER': ('site', 'twitter', ''),
        'SITE_FACEBOOK': ('site', 'facebook', ''),
        'FACEBOOK_APP_ID': ('site', 'facebook_app_id', ''),

        'BUILD_FOLDER': ('folders', 'build', 'build'),
        'PAGES_FOLDER': ('folders', 'pages', 'pages'),
        'POSTS_FOLDER': ('folders', 'posts', 'posts'),
        'STATIC_FOLDER': ('folders', 'static', 'static'),
        'TEMPLATE_FOLDER': ('folders', 'templates', 'templates'),

        'POSTS_EXTENSION': ('posts', 'extension', '.md'),

        'PRETTY_HTML': ('html', 'pretty', False),
        'MINIFY_HTML': ('html', 'minify', False),

        'SHOW_AUTHOR': ('author', 'show', True),
        'DEFAULT_AUTHOR': ('author', 'default_name', ''),
        'DEFAULT_AUTHOR_TWITTER': ('author', 'default_twitter', ''),
        'DEFAULT_AUTHOR_FACEBOOK': ('author', 'default_facebook', ''),
    }
    # Update app.config using the configuration keys
    for flask_key, (table, key, default) in config_keys.items():
        app.config[flask_key] = htmd_config.get(table, {}).get(key, default)
    app.static_folder = project_dir / app.config['STATIC_FOLDER']
    assert app.static_folder is not None
    # To avoid full paths in config.toml
    app.config['FLATPAGES_ROOT'] = (
        project_dir / app.config['POSTS_FOLDER']
    )
    app.config['FREEZER_DESTINATION'] = (
        project_dir / app.config['BUILD_FOLDER']
    )
    app.config['FREEZER_REMOVE_EXTRA_FILES'] = True
    # Allow build to be version controlled
    app.config['FREEZER_DESTINATION_IGNORE'] = ['.git*', '.hg*']
    app.config['FLATPAGES_EXTENSION'] = app.config['POSTS_EXTENSION']

    if Path(app.static_folder).is_dir():
        app.config['INCLUDE_CSS'] = (
            Path(app.static_folder) / 'combined.min.css').exists()
        app.config['INCLUDE_JS'] = (
            Path(app.static_folder) / 'combined.min.js').exists()
    else:
        # During testing they can be True from a previous test
        app.config['INCLUDE_CSS'] = False
        app.config['INCLUDE_JS'] = False

    # Without clearing the cache tests will use templates from the first test
    # Even when the template folder and jinja_loader has changed
    app.jinja_env.cache = {}

    # Allow config settings (even new user created ones) to be used in templates
    for key in app.config:
        app.jinja_env.globals[key] = app.config[key]

    favicon_path = Path(app.static_folder) / 'favicon.svg'
    app.jinja_env.globals['INCLUDE_DEFAULT_FAVICON'] = favicon_path.is_file()

    pages.template_folder = project_dir / app.config['PAGES_FOLDER']

    # Use templates in user set template folder
    # but fallback to app.template_folder (example_site/templates/)
    assert app.jinja_loader is not None
    app.jinja_loader = ChoiceLoader([
        FileSystemLoader(project_dir / app.config['TEMPLATE_FOLDER']),
        # Setting pages.template_folder is not enough
        FileSystemLoader(pages.template_folder),
        app.jinja_loader,
    ])

    # Without .reload() posts are from first test
    reload_posts(show_drafts)

    freezer.init_app(app)

    return app


def truncate_post_html(post_html: str) -> str:
    return BeautifulSoup(post_html[:255], 'html.parser').prettify()
app.jinja_env.globals['truncate_post_html'] = truncate_post_html

MONTHS = {
    '01': 'January',
    '02': 'February',
    '03': 'March',
    '04': 'April',
    '05': 'May',
    '06': 'June',
    '07': 'July',
    '08': 'August',
    '09': 'September',
    '10': 'October',
    '11': 'November',
    '12': 'December',
}


@app.after_request
def format_html(response: Response) -> ResponseReturnValue:
    if response.mimetype == 'text/html':
        if app.config.get('PRETTY_HTML', False):
            response.data = BeautifulSoup(
                response.data,
                'html.parser',
            ).prettify()
        elif app.config.get('MINIFY_HTML', False):
            response.data = minify(response.data.decode('utf-8'))
    return response


@pages.route('/<path:path>/')
def page(path: str) -> ResponseReturnValue:
    # ensure page is from pages directory
    # otherwise this will load any templates in the template folder
    pages_folder = pages.template_folder
    if not isinstance(pages_folder, Path) or not pages_folder.is_dir():
        abort(404)
    for page_path in pages_folder.iterdir():
        if path == page_path.stem:
            break
    else:
        abort(404)
    ret = render_template(path + '.html', active=path)
    return ret


app.register_blueprint(pages)


# Will end up in the static directory
@app.route('/static/pygments.css')
def pygments_css() -> ResponseReturnValue:
    return pygments_style_defs('tango'), 200, {'Content-Type': 'text/css'}


@app.route('/')
def index() -> ResponseReturnValue:
    latest = sorted(
        posts.published_posts,
        reverse=True,
        key=lambda p: p.meta['published'],
    )
    return render_template('index.html', active='home', posts=latest[:4])


@app.route('/feed.atom')
def feed() -> ResponseReturnValue:
    name = app.config.get('SITE_NAME')
    subtitle = app.config.get('SITE_DESCRIPTION') or 'Recent Blog Posts'
    url = app.config.get('URL')
    atom = AtomFeed(
        feed_url=url_for('all_posts'),
        subtitle=subtitle,
        title=name,
        url=url,
    )
    for post in posts.published_posts:
        url = url_for(
            'post',
            year=post.meta['published'].strftime('%Y'),
            month=post.meta['published'].strftime('%m'),
            day=post.meta['published'].strftime('%d'),
            path=post.path,
        )

        post_datetime = post.meta.get('updated') or post.meta['published']
        atom.add(
            post.meta.get('title'),
            post.html,
            content_type='html',
            author=post.meta.get('author', app.config.get('DEFAULT_AUTHOR')),
            url=url,
            updated=post_datetime,
        )
    ret = atom.get_response()
    return ret


@app.route('/all/')
def all_posts() -> ResponseReturnValue:
    latest = sorted(
        posts.published_posts,
        reverse=True,
        key=lambda p: p.meta['published'],
    )
    return render_template('all_posts.html', active='posts', posts=latest)


def draft_and_not_shown(post: Page) -> bool:
    is_draft = 'draft' in post.meta
    return is_draft and not posts.show_drafts and 'build' not in str(post.meta['draft'])


# If month and day are ints then Flask removes leading zeros
@app.route('/<year>/<month>/<day>/<path:path>/')
def post(year: str, month: str, day: str, path: str) -> ResponseReturnValue:
    if len(year) != 4 or len(month) != 2 or len(day) != 2:  # noqa: PLR2004
        abort(404)
    post = posts.get_or_404(path)
    if draft_and_not_shown(post):
        abort(404)
    date_str = f'{year}-{month}-{day}'
    if post.meta.get('published').strftime('%Y-%m-%d') != date_str:
        abort(404)
    return render_template('post.html', post=post)


@app.route('/draft/<post_uuid>/')  # noqa: RET503
def draft(post_uuid: str) -> ResponseReturnValue:
    for post in posts:
        if str(post.meta.get('draft', '')).replace('build|', '') == post_uuid:
            return render_template('post.html', post=post)
    abort(404)


@app.route('/tags/')
def all_tags() -> ResponseReturnValue:
    tag_counts: dict[str, int] = {}
    for post in posts.published_posts:
        for tag in post.meta.get('tags', []):
            if tag not in tag_counts:
                tag_counts[tag] = 0
            tag_counts[tag] += 1
    return render_template('all_tags.html', active='tags', tags=tag_counts)


def no_posts_shown(post_list: list[Page]) -> bool:
    return all(
        'draft' in p.meta and 'build' not in str(p.meta['draft'])
        for p in post_list
    )

@app.route('/tags/<string:tag>/')
def tag(tag: str) -> ResponseReturnValue:
    tagged = [p for p in posts if tag in p.meta.get('tags', [])]
    if not tagged:
        abort(404)
    if not posts.show_drafts and no_posts_shown(tagged):
        abort(404)
    if posts.show_drafts:
        tagged_published = tagged
    else:
        tagged_published = [p for p in tagged if 'draft' not in p.meta]
    sorted_posts = sorted(
        tagged_published,
        reverse=True,
        key=lambda p: p.meta.get('published'),
    )
    return render_template('tag.html', posts=sorted_posts, tag=tag)


@app.route('/author/<author>/')
def author(author: str) -> ResponseReturnValue:
    # if the author has a draft build
    # page is served without displaying posts
    # so no 404 when for the link from the draft
    posts_author = [p for p in posts if author == p.meta.get('author', '')]

    if not posts_author:
        abort(404)

    if not posts.show_drafts and no_posts_shown(posts_author):
        abort(404)
    if posts.show_drafts:
        posts_author_published = posts_author
    else:
        posts_author_published = [p for p in posts_author if 'draft' not in p.meta]

    posts_sorted = sorted(
        posts_author_published,
        reverse=True,
        key=lambda p: p.meta.get('published'),
    )
    return render_template(
        'author.html',
        active='author',
        author=author,
        posts=posts_sorted,
    )


@app.route('/404.html')
def not_found() -> ResponseReturnValue:
    return render_template('404.html')


@app.route('/<int:year>/')
def year_view(year: int) -> ResponseReturnValue:
    year_str = str(year)
    if len(year_str) != len('YYYY'):
        abort(404)
    year_posts = [
        p for p in posts.published_posts
        if year_str == p.meta['published'].strftime('%Y')
    ]
    if not year_posts:
        abort(404)
    sorted_posts = sorted(
        year_posts,
        reverse=False,
        key=lambda p: p.meta['published'],
    )
    return render_template('year.html', year=year_str, posts=sorted_posts)


@app.route('/<year>/<month>/')
def month_view(year: str, month: str) -> ResponseReturnValue:
    month_posts = [
        p for p in posts.published_posts if year == p.meta['published'].strftime('%Y')
        and month == p.meta['published'].strftime('%m')
    ]
    if not month_posts:
        abort(404)
    sorted_posts = sorted(
        month_posts,
        reverse=False,
        key=lambda p: p.meta['published'],
    )
    month_string = MONTHS[month]
    return render_template(
        'month.html',
        year=year,
        month_string=month_string,
        posts=sorted_posts,
    )


@app.route('/<year>/<month>/<day>/')
def day_view(year: str, month: str, day: str) -> ResponseReturnValue:
    day_posts = [
        p for p in posts.published_posts if year == p.meta['published'].strftime('%Y')
        and month == p.meta['published'].strftime('%m')
        and day == p.meta['published'].strftime('%d')
    ]
    if not day_posts:
        abort(404)
    month_string = MONTHS[month]
    return render_template(
        'day.html',
        year=year,
        month_string=month_string,
        day=day,
        posts=day_posts,
    )


@app.errorhandler(404)
def page_not_found(_e: Exception | int) -> ResponseReturnValue:
    return render_template('404.html'), 404


# Telling Frozen-Flask about routes that are not linked to in templates
@freezer.register_generator  # type: ignore[no-redef]
def year_view() -> Iterator[dict]:  # noqa: F811
    for post in posts.published_posts:
        yield {
            'year': post.meta['published'].year,
        }


@freezer.register_generator  # type: ignore[no-redef]
def month_view() -> Iterator[dict]:  # noqa: F811
    for post in posts.published_posts:
        yield {
            'month': post.meta['published'].strftime('%m'),
            'year': post.meta['published'].year,
        }


@freezer.register_generator  # type: ignore[no-redef]
def day_view() -> Iterator[dict]:  # noqa: F811
    for post in posts.published_posts:
        yield {
            'day': post.meta['published'].strftime('%d'),
            'month': post.meta['published'].strftime('%m'),
            'year': post.meta['published'].year,
        }


@freezer.register_generator  # type: ignore[no-redef]
def draft() -> Iterator[dict]:  # noqa: F811
    draft_posts = [
        p for p in posts
        if 'draft' in p.meta and 'build' in str(p.meta['draft'])
    ]
    for post in draft_posts:
        if not valid_uuid(post.meta['draft'].replace('build|', '')):
            post.meta['draft'] = 'build|' + str(uuid.uuid4())
            set_post_metadata(app, post, 'draft', post.meta['draft'])
        yield {
            'post_uuid': post.meta['draft'].replace('build|', ''),
        }


@freezer.register_generator  # type: ignore[no-redef]
def page() -> Iterator[str]:  # noqa: F811
    pages_folder = pages.template_folder
    if not isinstance(pages_folder, Path) or not pages_folder.is_dir():
        return
    for page in pages_folder.iterdir():
        # Need to create for pages.page
        # Since this route is in a different Blueprint
        # Using the URL works
        yield f'/{page.stem}/'



from importlib.resources import as_file, files
from pathlib import Path
import shutil
import uuid

import click
from csscompressor import compress
from flask import Flask
from flask_flatpages import Page
from jsmin import jsmin


def create_directory(name: str) -> Path:
    directory = Path(name)
    try:
        directory.mkdir()
    except FileExistsError:
        msg = f'{name} already exists and was not created.'
        click.echo(click.style(msg, fg='yellow'))
    else:
        click.echo(click.style(f'{name} was created.', fg='green'))
    return directory


def combine_and_minify_css(static_folder: Path) -> bool:
    # Combine and minify all .css files in the static folder
    try:
        css_files = sorted([
            f for f in static_folder.iterdir()
            if f.name.endswith('.css') and f.name != 'combined.min.css'
        ])
    except FileNotFoundError:
        # static folder does not exist
        return False

    if not css_files:
        # There are no .css files in the static folder
        return False

    # combine all .css files into one string
    file_contents = []
    for f in css_files:
        with f.open('r') as css_file:
            file_contents.append(css_file.read())
    combined_str = '\n'.join(file_contents)

    try:
        with (static_folder / 'combined.min.css').open('r') as combined_file:
            current_combined = combined_file.read()
    except FileNotFoundError:
        current_combined = ''

    new_combined = compress(combined_str)
    if new_combined == current_combined:
        return False

    with (static_folder / 'combined.min.css').open('w') as combined_file:
        combined_file.write(new_combined)
    return True


def combine_and_minify_js(static_folder: Path) -> bool:
    # Combine and minify all .js files in the static folder
    try:
        js_files = sorted([
            f for f in static_folder.iterdir()
            if f.name.endswith('.js')
            and f.name != 'combined.min.js'
        ])
    except FileNotFoundError:
        # static folder does not exist
        return False

    if not js_files:
        # There are no .js files in the static folder
        return False

    # combine all .js files into one string
    file_contents = []
    for f in js_files:
        with f.open('r') as js_file:
            file_contents.append(js_file.read())
    combined_str = '\n'.join(file_contents)

    try:
        with (static_folder / 'combined.min.js').open('r') as combined_file:
            current_combined = combined_file.read()
    except FileNotFoundError:
        current_combined = ''

    new_combined = jsmin(combined_str)
    if new_combined == current_combined:
        return False

    with (static_folder / 'combined.min.js').open('w') as combined_file:
        combined_file.write(new_combined)
    return True


def copy_file(source: Path, destination: Path) -> None:
    if destination.exists() is False:
        shutil.copyfile(source, destination)
        click.echo(click.style(f'{destination} was created.', fg='green'))
    else:
        msg = f'{destination} already exists and was not created.'
        click.echo(click.style(msg, fg='yellow'))


def copy_missing_templates() -> None:
    template_dir = files('htmd.example_site') / 'templates'
    with as_file(template_dir) as template_path:
        for template_file in sorted(template_path.iterdir()):
            file_name = template_file.name
            copy_file(template_file, Path('templates') / file_name)


def copy_site_file(directory: Path, filename: str) -> None:
    if directory.name == '':
        anchor = 'htmd.example_site'
    else:
        anchor = f'htmd.example_site.{directory}'
    source_path = files(anchor) / filename
    destination_path = directory / filename

    with as_file(source_path) as file:
        copy_file(file, destination_path)


def set_post_metadata(
    app: Flask,
    post: Page,
    field: str,
    value: str,
) -> None:
    file_path = (
        Path(app.config['FLATPAGES_ROOT'])
        / (post.path + app.config['FLATPAGES_EXTENSION'])
    )
    with file_path.open('r') as file:
        lines = file.readlines()

    found = False
    with file_path.open('w') as file:
        for line in lines:
            if not found and field  + ':' in line:
                file.write(f'{field}: {value}\n')
                found = True
            elif not found and line == '...\n':
                # Write field and value before '...'
                file.write(f'{field}: {value}\n')
                file.write(line)
                found = True
            else:
                file.write(line)


def valid_uuid(string: str) -> bool:
    try:
        uuid.UUID(string, version=4)
    except ValueError:
        return False
    else:
        return True


def send_stderr(message: str) -> None:
    click.echo(click.style(message, fg='red'), err=True)


def validate_post(post: Page, required_fields: list[str]) -> bool: # noqa: C901
    correct = True
    for field in required_fields:
        if field not in post.meta:
            correct = False
            msg = f'Post "{post.path}" does not have field {field}.'
            send_stderr(msg)
    if 'published' in post.meta:
        published = post.meta.get('published')
        if not hasattr(published, 'year'):
            correct = False
            msg = (
                f'Published date {published} for {post.path}'
                ' is not in the format YYYY-MM-DD.'
            )
            send_stderr(msg)
    if 'updated' in post.meta:
        updated = post.meta.get('updated')
        if not hasattr(updated, 'year'):
            correct = False
            msg = (
                f'Updated date {updated} for {post.path}'
                ' is not in the format YYYY-MM-DD.'
            )
            send_stderr(msg)
    if 'draft' in post.meta:
        draft = post.meta['draft']
        if draft in [True, False, 'build']:
            pass
        elif 'build|' in draft:
            draft_id = draft.split('|')[1]
            if not valid_uuid(draft_id):
                correct = False
                msg = (
                    f'Draft field {draft} for {post.path}'
                    ' has an invalid UUID4.'
                )
                send_stderr(msg)
        else:
            correct = False
            msg = (
                f'Draft field {draft} for {post.path}'
                ' is not valid. It must be True, False,'
                ' "build", or "build|<UUID4>".'
            )
            send_stderr(msg)

    return correct
